<?php

namespace Drupal\entity_reference_layout\Plugin\Field\FieldWidget;

use Drupal\file\Entity\File;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\inline_entity_form\Plugin\Field\FieldWidget\InlineEntityFormComplex;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\entity_reference_layout\Event\ErlPropertiesFormEvent;
use Drupal\Core\Layout\LayoutPluginManager;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Session\AccountProxy;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Entity Reference with Layout field widget.
 *
 * @FieldWidget(
 *   id = "entity_reference_layout_widget",
 *   label = @Translation("Entity reference layout (With layout builder)"),
 *   description = @Translation("Layout builder for paragraphs."),
 *   field_types = {
 *     "entity_reference_layout_revisioned"
 *   },
 *   multiple_values = true
 * )
 */
class EntityReferenceLayoutWidget extends InlineEntityFormComplex {

  /**
   * The List of allowed Layouts.
   *
   * @var array
   */
  protected $allowedLayouts;

  /**
   * The Layouts Manager.
   *
   * @var \Drupal\Core\Layout\LayoutPluginManager
   */
  protected $layoutPluginManager;

  /**
   * The Event Dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The Renderer service property.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The Current User.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Constructs an EntityReferenceLayoutWidget object.
   *
   * @param string $plugin_id
   *   The plugin_id for the widget.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the widget is associated.
   * @param array $settings
   *   The widget settings.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle info.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   The entity display repository.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   Module handler service.
   * @param \Drupal\Core\Layout\LayoutPluginManager $layoutPluginManager
   *   Layout plugin manager service.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $eventDispatcher
   *   Event dispatcher service.
   * @param \Drupal\Core\Render\Renderer $renderer
   *   Renderer service.
   * @param \Drupal\Core\Session\AccountProxy $current_user
   *   Current logged in user.
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    array $third_party_settings,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
    EntityTypeManagerInterface $entity_type_manager,
    EntityDisplayRepositoryInterface $entity_display_repository,
    ModuleHandlerInterface $module_handler,
    LayoutPluginManager $layoutPluginManager,
    EventDispatcherInterface $eventDispatcher,
    Renderer $renderer,
    AccountProxy $current_user) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings, $entity_type_bundle_info, $entity_type_manager, $entity_display_repository, $module_handler);
    $this->layoutPluginManager = $layoutPluginManager;
    $this->eventDispatcher = $eventDispatcher;
    $this->renderer = $renderer;
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings'],
      $container->get('entity_type.bundle.info'),
      $container->get('entity_type.manager'),
      $container->get('entity_display.repository'),
      $container->get('module_handler'),
      $container->get('plugin.manager.core.layout'),
      $container->get('event_dispatcher'),
      $container->get('renderer'),
      $container->get('current_user')
    );
  }

  /**
   * Prepare the form state.
   *
   * First calls parent method, then loads
   * layout info into $form_state from items.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   *   The field values.
   * @param bool $translating
   *   Whether there's a translation in progress.
   */
  protected function prepareFormState(FormStateInterface $form_state, FieldItemListInterface $items, $translating = FALSE) {

    parent::prepareFormState($form_state, $items, $translating);
    $entities = $form_state->get([
      'inline_entity_form',
      $this->getIefId(),
      'entities',
    ]);

    foreach ($items as $delta => $item) {
      if (isset($entities[$delta]) && !isset($entities[$delta]['layout'])) {
        $entities[$delta]['layout']['options'] = $item->options;
        $entities[$delta]['layout']['region'] = isset($item->region) ? $item->region : '';
        $entities[$delta]['layout']['layout'] = isset($item->layout) ? $item->layout : '';
        $entities[$delta]['layout']['entity_bundle'] = $item->entity->bundle();
        $entities[$delta]['layout']['parent'] = -1;
      }
    }
    $form_state->set(['inline_entity_form', $this->getIefId(), 'entities'], $entities);
  }

  /**
   * Stores new or edited entity in $form_state.
   *
   * Stored temporarily in $form_state, saved to db
   * when container entity is saved down the line.
   *
   * Except for the lines that save layout info with entity,
   * this is copied directly from the InlineEntityFormComplex class.
   */
  public static function submitSaveEntity($entity_form, FormStateInterface $form_state) {

    $ief_id = $entity_form['#ief_id'];

    /** @var \Drupal\Core\Entity\EntityInterface $entity */
    $entity = $entity_form['#entity'];

    // Creating a new entity.
    if (in_array($entity_form['#op'], ['add', 'duplicate'])) {

      $layout_options = $form_state->getValue(array_merge($entity_form['#parents'], ['layout_options']));
      $layout_properties = $form_state->getValue(array_merge($entity_form['#parents'], ['layout_properties']));

      // The "layout" value is really a property, but exists in the
      // options form so a user can choose the desired layout explicity.
      // Move it out of options:
      $layout_name = '';
      if (isset($layout_options['layout'])) {
        $layout_name = $layout_options['layout'];
        unset($layout_options['layout']);
      }
      $layout = [
        'layout' => $layout_name,
        'region' => $layout_properties['region'],
        'parent' => $layout_properties['parent'],
        'options' => $layout_options,
        'entity_bundle' => $entity->bundle(),
      ];

      // Determine the correct weight of the new element.
      $weight = 0;
      $entities = $form_state->get(['inline_entity_form', $ief_id, 'entities']);
      if (!empty($entities)) {
        $weight = max(array_keys($entities)) + 1;
      }
      // Add the entity to form state, mark it for saving, and close the form.
      $entities[] = [
        'entity' => $entity,
        'weight' => $weight,
        'form' => NULL,
        'needs_save' => TRUE,
        'layout' => $layout,
      ];
      $form_state->set(['inline_entity_form', $ief_id, 'entities'], $entities);
    }

    // Editing an existing entity.
    else {

      $parents = array_merge($entity_form['#parents'], ['layout_options']);
      $layout_options = $form_state->getValue($parents);

      // When editing entities, properties other than the chosen layout
      // are manipulated separately -- not in this element of the form.
      $layout_name = '';
      if (isset($layout_options['layout'])) {
        $layout_name = $layout_options['layout'];
        unset($layout_options['layout']);
      }

      $delta = $entity_form['#ief_row_delta'];
      $entities = $form_state->get(['inline_entity_form', $ief_id, 'entities']);
      $entities[$delta]['entity'] = $entity;
      $entities[$delta]['needs_save'] = TRUE;
      $entities[$delta]['layout']['options'] = $layout_options;
      $entities[$delta]['layout']['layout'] = $layout_name;
      $form_state->set(['inline_entity_form', $ief_id, 'entities'], $entities);
    }
  }

  /**
   * Build the form widget.
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {

    $element = parent::formElement($items, $delta, $element, $form, $form_state);

    $element['entities']['#theme'] = 'entity_reference_layout_widget';
    $entities = $form_state->get([
      'inline_entity_form',
      $this->getIefId(),
      'entities',
    ]);
    $element['entities']['#widget'] = 'entity_reference_layout_widget';

    $handler_settings = $items->getSetting('handler_settings');
    $layout_bundles = $element['entities']['#layout_bundles'] = $handler_settings['layout_bundles'];
    $this->allowedLayouts = isset($handler_settings['allowed_layouts']) ? $handler_settings['allowed_layouts'] : [];

    if (!empty($handler_settings['icons'])) {
      foreach ($handler_settings['icons'] as $name => $value) {
        if (!empty($value) && !empty($value[0])) {
          if ($file = File::load($value[0])) {
            $element['entities']['#attached']['drupalSettings']['erlIcons'][$name] = $file->url();
          }
        }
      }
    }

    $field_name = $this->fieldDefinition->getName();
    $element['#field_name'] = $field_name;

    if (isset($element['actions'])) {

      $element['actions']['#attributes']['class'][] = 'erl-field-actions';
      $element['actions']['bundle']['#attributes']['class'][] = 'hidden';
      $element['actions']['ief_add']['#attributes']['style'][] = 'display:none;';
      if (!count($entities)) {
        $element['actions']['#attributes']['class'][] = 'empty';
      }
      $nested_bundles = ['Layout' => [], 'Content' => []];
      foreach ($element['actions']['bundle']['#options'] as $key => $val) {
        $bundle_type = in_array($key, $layout_bundles) ? 'Layout' : 'Content';
        $nested_bundles[$bundle_type][$key] = $val;
      }
      $element['actions']['bundle']['#options'] = $nested_bundles;
      $element['actions']['region'] = [
        '#type' => 'textfield',
        '#placeholder' => 'Region',
        '#size' => 15,
        '#attributes' => [
          'class' => ['erl-new-item-region', 'hidden'],
        ],
      ];
      $element['actions']['delta'] = [
        '#type' => 'textfield',
        '#placeholder' => 'Delta',
        '#size' => 15,
        '#attributes' => [
          'class' => ['erl-new-item-delta', 'hidden'],
        ],
      ];
    }

    // If we're forcing users to use layouts, and no content has been added,
    // only show options for layout bundles.
    if ($this->getSetting('force_layouts') && empty($entities)) {
      unset($element['actions']['bundle']['#options']['Content']);
    }

    if ($this->getSetting('full_screen_editing')) {
      $element['entities']['#attached']['library'][] = 'entity_reference_layout/full_screen_editing';
    }

    foreach ($entities as $delta => $value) {

      $defaults = $value['layout'];

      // We set #layout_parent to instruct the theme function where to position
      // new entities that have not been saved to the db yet but need to be
      // represented on the screen, in right place.
      if (isset($defaults['parent']) && is_numeric($defaults['parent'])) {
        $element['entities'][$delta]['#layout_parent'] = $defaults['parent'];
      }
      $defaults['is_layout_bundle'] = in_array($defaults['entity_bundle'], $layout_bundles);

      /* @var \Drupal\core\Entity\EntityInterface $entity */
      $entity = $value['entity'];
      $form_item =& $element['entities'][$delta];

      $view_builder = $this->entityTypeManager->getViewBuilder($entity->getEntityTypeId());
      $preview = $view_builder->view($entity);
      $rendered_preview = $this->renderer->render($preview);
      $form_item['prefix'] = [
        '#markup' => $rendered_preview,
        '#weight' => -100,
      ];
      if (!empty($form_item['form'])) {
        if ($value['form'] == 'remove') {
          try {
            $form_item['form']['#attributes']['class'][] = 'erl-remove-form';
            /* @var \Drupal\core\Entity\EntityInterface $form_item_entity */
            $form_item_entity = $form_item['#entity'];
            $bundle_type_id = $form_item_entity->getEntityType()
              ->getBundleEntityType();
            $bundle_label = $this->entityTypeManager->getStorage($bundle_type_id)
              ->load($form_item_entity->bundle())
              ->label();
            if ($bundle_label) {
              $form_item['form']['message']['#markup'] = $this->t('Are you sure you want to remove this %bundle?', ['%bundle' => $bundle_label]);
              $form_item['form']['message']['#attributes']['class'] = ['erl-remove-message'];
            }
          }
          catch (\Exception $e) {
            watchdog_exception('entity_reference_layout', $e);
          }
        }
      }
      if (!empty($form_item['form']['inline_entity_form'])) {
        $defaults['hidden'] = FALSE;
        // Allow module to alter forms.
        $form_item['form']['#process'][] = 'entity_reference_layout_alter_form';
        $form_item['form']['#theme'] = 'entity_reference_layout_ief';
        $form_item['form']['inline_entity_form']['#process'][] = [get_class($this), 'alterEntityFormActions'];
        $form_item['form']['inline_entity_form'] += $this->layoutOptionsForm($defaults);
        $form_item['form']['inline_entity_form']['#erl_form'] = TRUE;
      }
      else {
        $defaults['hidden'] = TRUE;
      }
      $form_item += $this->layoutPropertiesForm($defaults);

      // Hide delta without relying on javascript.
      $form_item['delta']['#attributes']['class'][] = 'hidden';

      // Set layout info for rendering in theme function.
      if (in_array($entity->bundle(), $layout_bundles) && !empty($defaults['layout'])) {
        $form_item['#layout'] = $defaults['layout'];
      }
      elseif (!empty($defaults['region'])) {
        $form_item['#region'] = $defaults['region'];
      }
      if (!empty($value['layout']['options'])) {
        $form_item['#layout_options'] = $value['layout']['options'];
      }

    }

    // Add layout options for inline entity form:
    if (isset($element['form']['inline_entity_form'])) {

      // Allow module to alter form.
      $element['form']['#theme'] = 'entity_reference_layout_ief';
      $element['form']['#process'][] = 'entity_reference_layout_alter_form';
      $element['form']['#attributes']['class'][] = 'erl-add-entity-form';
      $element['form']['inline_entity_form']['#erl_form'] = TRUE;
      $element['form']['inline_entity_form']['#process'][] = [get_class($this), 'alterEntityFormActions'];

      // Using raw input because $form_state values
      // are not dependable at this point.
      // IMPORTANT: $input values MUST BE sanitized manually below.
      $input = $form_state->getUserInput();
      $defaults = [
        'parent' => 10000,
        'entity_bundle' => $element['form']['inline_entity_form']['#bundle'],
        'is_new' => TRUE,
      ];
      $defaults['is_layout_bundle'] = in_array($defaults['entity_bundle'], $layout_bundles);
      if (isset($input[$field_name]['actions']['region'])) {
        $defaults['region'] = preg_replace('/[^A-Za-z0-9\-\_]/', '', $input[$field_name]['actions']['region']);
      }
      if (isset($input[$field_name]['actions']['delta'])) {
        $defaults['parent'] = intval($input[$field_name]['actions']['delta']);
      }
      $element['form']['inline_entity_form'] += $this->layoutPropertiesForm($defaults);
      $element['form']['inline_entity_form'] += $this->layoutOptionsForm($defaults);
    }

    // Add validation to force content inside regions, if applicable.
    if ($this->getSetting('force_layouts')) {
      $element['entities']['#force_layouts'] = TRUE;
    }
    else {
      $element['entities']['#force_layouts'] = FALSE;
    }
    $element['#attributes'] = [
      'class' => ['erl-field'],
    ];
    $element['#attached']['library'][] = 'core/drupal.dialog';
    return $element;
  }

  /**
   * Manipulate form actions form element.
   */
  public static function alterEntityFormActions($element) {
    $element['actions']['#weight'] = 10000;
    $element['actions']['#attributes']['class'][] = 'erl-form-actions';

    if (!empty($element['#bundle'])) {
      try {
        /* @var \Drupal\core\Entity\EntityInterface $element_entity */
        $element_entity = $element['#entity'];
        $bundle_type_id = $element_entity->getEntityType()
          ->getBundleEntityType();
        $bundle_label = \Drupal::entityTypeManager()
          ->getStorage($bundle_type_id)
          ->load($element['#bundle'])
          ->label();

        if ($element['#op'] == 'add') {
          $save_label = t('Create @type_singular', ['@type_singular' => $bundle_label]);
        }
        elseif ($element['#op'] == 'duplicate') {
          $save_label = t('Duplicate @type_singular', ['@type_singular' => $bundle_label]);
        }
        else {
          $save_label = t('Update @type_singular', ['@type_singular' => $bundle_label]);
        }
        $element['actions']['ief_' . $element['#op'] . '_save']['#value'] = $save_label;
      }
      catch (\Exception $e) {
        watchdog_exception('entity_reference_layout', $e);
      }
    }
    return $element;
  }

  /**
   * Ensure referenced entities are assigned to a region.
   */
  public static function validateContentInRegions($element, FormStateInterface $form_state) {
    $layout_bundles = $element['entities']['#layout_bundles'];
    $field_name = $element['#field_name'];
    if ($form_state->getValue([$field_name, 'entities'])) {
      foreach ($form_state->getValue([$field_name, 'entities']) as $form_item) {
        if (!in_array($form_item['layout']['entity_bundle'], $layout_bundles)) {
          if (empty($form_item['layout']['region'])) {
            $form_state->setError($element, t("Content needs to be within regions."));
            return;
          }
        }
      }
    }
  }

  /**
   * Build layout properties form.
   *
   * @todo: Use configuration options from layout plugins.
   *   see https://www.drupal.org/project/entity_reference_layout/issues/3038533
   */
  private function layoutPropertiesForm($defaults) {

    $form_item = [];
    $form_item['layout_properties'] = [
      '#type' => 'container',
      '#weight' => -1000,
    ];
    $form_item['layout_properties']['#attributes'] = ['class' => ['hidden']];
    $form_item['layout_properties']['layout'] = [
      '#type' => 'hidden',
      // Use value because this doesn't change from this form element.
      '#value' => isset($defaults['layout']) ? $defaults['layout'] : '',
      '#attributes' => [
        'class' => [
          'erl-layout-select',
          'erl-layout',
          'erl-layout-' . (!empty($defaults['layout']) ? $defaults['layout'] : ''),
        ],
      ],
    ];
    $form_item['layout_properties']['region'] = [
      '#type' => 'hidden',
      '#default_value' => $defaults['region'],
      '#attributes' => [
        'class' => [
          'erl-region-select',
          'erl-region',
          'erl-region-' . (!empty($defaults['region']) ? $defaults['region'] : ''),
        ],
      ],
    ];
    $form_item['layout_properties']['entity_bundle'] = [
      '#type' => 'value',
      '#value' => $defaults['entity_bundle'],
    ];
    $form_item['layout_properties']['parent'] = [
      '#type' => 'hidden',
      '#default_value' => intval($defaults['parent']),
      '#attributes' => [
        'class' => ['parent-delta'],
      ],
    ];

    return $form_item;
  }

  /**
   * Add theme wrappers to layout selection radios.
   */
  public static function processLayoutOptions($element) {
    foreach (Element::children($element) as $radio_item) {
      $element[$radio_item]['#theme_wrappers'][] = 'entity_reference_layout_radio';
    }
    return $element;
  }

  /**
   * Build a layout options form.
   *
   * @todo: Use form from layout plugins.
   *   See https://www.drupal.org/project/entity_reference_layout/issues/3038533
   */
  private function layoutOptionsForm($defaults) {

    $form_item = [];
    $form_item['layout_options'] = [
      '#type' => 'container',
      '#weight' => $defaults['is_layout_bundle'] ? -1000 : 1000,
    ];

    if ($defaults['is_layout_bundle']) {
      $layout_options = [];
      foreach ($this->allowedLayouts as $layout_group) {
        foreach ($layout_group as $layout_id => $layout_name) {
          $layout_options[$layout_id] = $layout_name;
        }
      }
      $form_item['layout_options']['layout'] = [
        '#type' => 'radios',
        '#title' => $this->t('Select a layout:'),
        '#description' => $this->t('Choose a layout for this section.'),
        '#options' => $layout_options,
        '#default_value' => !empty($defaults['layout']) ? $defaults['layout'] : key($layout_options),
        '#attributes' => [
          'class' => ['erl-layout', 'erl-layout-section-select'],
        ],
        '#required' => TRUE,
        '#after_build' => [[get_class($this), 'processLayoutOptions']],
      ];
      $form_item['layout_options']['options'] = [
        '#type' => 'details',
        '#title' => $this->t('Layout Options'),
        '#description' => $this->t('Classes will be applied to the container for this field item.'),
        '#open' => FALSE,
        '#access' => $this->currentUser->hasPermission('adjust entity reference layout options'),
      ];
      $form_item['layout_options']['options']['container_classes'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Custom Classes for Layout Container'),
        '#description' => $this->t('Classes will be applied to the container for this field item.'),
        '#size' => 20,
        '#default_value' => !empty($defaults['options']['options']['container_classes']) ? $defaults['options']['options']['container_classes'] : '',
        '#placeholder' => $this->t('CSS Classes'),
      ];
      $form_item['layout_options']['options']['bg_color'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Background Color for Layout Container'),
        '#description' => $this->t('Background will be applied to the layout container.'),
        '#size' => 10,
        '#default_value' => !empty($defaults['options']['options']['bg_color']) ? $defaults['options']['options']['bg_color'] : '',
        '#placeholder' => $this->t('Hex Code'),
      ];
      $event = new ErlPropertiesFormEvent($form_item, $defaults);
      $this->eventDispatcher->dispatch(ErlPropertiesFormEvent::EVENT_NAME, $event);
    }
    return $form_item;

  }

  /**
   * Field instance settings form.
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form = parent::settingsForm($form, $form_state);

    // @FIXME
    // Allow existing doesn't work at the moment (it breaks all the widget).
    // It is thus disabled at the moment.
    $form['allow_existing']['#disabled'] = TRUE;
    $form['allow_existing']['#description'] = $this->t('This feature is not working at the moment (disabled)');

    // @FIXME
    // Allow duplicate doesn't work at the moment (unpredictable results).
    // correct behavior. It is thus disabled at the moment.
    $form['allow_duplicate']['#disabled'] = TRUE;
    $form['allow_duplicate']['#description'] = $this->t('This feature is not working at the moment (disabled)');

    $form['full_screen_editing'] = [
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('full_screen_editing'),
      '#title' => $this->t('Full screen editing'),
      '#description' => $this->t('Show node edit form full-screen to make room for layout builder'),
    ];
    $form['force_layouts'] = [
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('force_layouts'),
      '#title' => $this->t('Force layouts'),
      '#description' => $this->t('Forces all added content to be created within a layout region.'),
    ];
    return $form;
  }

  /**
   * Default settings for widget.
   */
  public static function defaultSettings() {
    $defaults = parent::defaultSettings();
    $defaults += [
      'full_screen_editing' => FALSE,
      'force_layouts' => FALSE,
    ];

    return $defaults;
  }

  /**
   * Massage settings values for use in form.
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {

    $field_name = $this->fieldDefinition->getName();
    $field_value = $form_state->getValue($field_name);
    $entities = $form_state->get([
      'inline_entity_form',
      $this->getIefId(),
      'entities',
    ]);

    foreach ($values as $delta => $value) {
      if (isset($field_value['entities'][$delta])) {
        $values[$delta]['region'] = !empty($field_value['entities'][$delta]['layout_properties']['region']) ? $field_value['entities'][$delta]['layout_properties']['region'] : '';
        $values[$delta]['layout'] = !empty($field_value['entities'][$delta]['layout_properties']['layout']) ? $field_value['entities'][$delta]['layout_properties']['layout'] : '';
      }
      // Other options come from stored entities.
      if (isset($entities[$delta]['layout']['options'])) {
        $values[$delta]['options'] = $entities[$delta]['layout']['options'];
      }
    }
    return $values;
  }

}
